/*\
 * MICE Column List View
 * Version 1.02
 * Copyright (C) 2000-2005 Colin Munro
 *
 * ListViewer.cpp - definitions of main objects for the CColumnListView object.
 *
 * Please see license.txt and readme.txt for more information on this product. Please
 * see readme.txt for more information if you wish to use this software in your own
 * product.
\*/
#define _BUILDINGLIB_
#include <memory.h>
#include <Application.h>
#include <Bitmap.h>
#include <Cursor.h>
#include <MenuItem.h>
#include <Message.h>
#include <PopUpMenu.h>
#include <Region.h>
#include <ScrollBar.h>
#include "ListViewer.h"
#include "ListViewerSort.h"

#include "ListViewerCursor.h"		// Contains mouse cursors used by the view

#define COLSEP				0
#define DRAGTHRESHOLD		6
#define COLUMNMOVE			'LVcl'
#define COLLEADIN			16
#define EXPANDERWIDTH		16

#define MULTISELECT()		(modifiers()&B_SHIFT_KEY)

class CColumnMenuItem : public BMenuItem
{
public:
	CColumnMenuItem(CColumn *column,char *name,int id)
		:BMenuItem(name,new BMessage(id))
		{
			this->column=column;
		};
	CColumn *column;
};

// CField

CField::CField()
{
	parent=NULL;
}
CField::~CField()
{
	if(parent!=NULL) {
		if(last!=NULL) last->next=next;
		if(next!=NULL) next->last=last;
		if(this==(*parent)) (*parent)=(next!=NULL)?next:last;
	}
}
void CField::Attached(CField **parent)
{
	this->parent=parent;
	last=NULL;
	next=(*parent);
	if(next!=NULL) next->last=this;
	(*parent)=this;
}

// CRow

CRow::CRow(float height)
{
	parent=NULL;
	children=NULL;
	bExpanded=false;
	fHeight=height;
	fields=NULL;
	owner=NULL;
	selection_last=selection_next=NULL;
}
CRow::~CRow()
{
	while(children!=NULL) delete children;
	while(fields!=NULL) delete fields;
	Unlink();
}
void CRow::Attached(CRow *owner,CRow **parent)
{
	this->owner=owner;
	this->parent=parent;
	last=NULL;
	next=(*parent);
	if(next!=NULL) next->last=this;
	(*parent)=this;
}
void CRow::Unlink(void)
{
	if(parent!=NULL) {
		if(last!=NULL) last->next=next;
		if(next!=NULL) next->last=last;
		if(this==(*parent)) (*parent)=(next!=NULL)?next:last;
		parent=NULL;
	}
}
void CRow::AddChild(CRow *child)
{
	child->Attached(this,&children);
}
void CRow::DrawExpander(BView *targetView,BRect bigrect)
{
	float w,h;
	BRect rect;

	if(IsExpandable()) {
		// Draw -
		w=(bigrect.right-bigrect.left)/2;
		h=(bigrect.bottom-bigrect.top)/2;
		rect.left=bigrect.left+(w/2);
		rect.top=bigrect.top+(h/2);
		rect.right=rect.left+w;
		rect.bottom=rect.top+h;
		targetView->SetHighColor(255,255,255);
		targetView->FillRect(rect);
		targetView->SetHighColor(0,0,0);
		targetView->StrokeRect(rect);
		targetView->StrokeLine(BPoint(rect.left+2,rect.top+((rect.bottom-rect.top)/2)),BPoint(rect.right-2,rect.top+((rect.bottom-rect.top)/2)));
		if(!IsExpanded())
			targetView->StrokeLine(BPoint(rect.left+((rect.right-rect.left)/2),rect.top+2),BPoint(rect.left+((rect.right-rect.left)/2),rect.bottom-2));
	}
}
bool CRow::DrawIcon(BView *targetView,BRect rect,float *width)
{
	return false;
}
bool CRow::IsPointWithinExpander(BRect bigrect,BPoint point)
{
	float w,h;
	BRect rect;
	
	if(IsExpandable()) {
		w=(bigrect.right-bigrect.left)/2;
		h=(bigrect.bottom-bigrect.top)/2;
		rect.left=bigrect.left+(w/2);
		rect.top=bigrect.top+(h/2);
		rect.right=rect.left+w;
		rect.bottom=rect.top+h;
		return rect.Contains(point);
	} else return false;
}
void CRow::ExpandOperation(bool bExpanding)
{
	bExpanded=bExpanding;
}
void CRow::ExpandOperationComplete(bool bExpanding)
{
	// Nothing to do
}
bool CRow::IsExpandable(void)
{
	return children!=NULL;
}
void CRow::AddField(CField *field,int iFieldIdentifier)
{
	field->Attached(&fields);
	field->iFieldIdentifier=iFieldIdentifier;
}
bool CRow::IsExpanded(void) const
{
	return bExpanded;
}
float CRow::Height(void) const
{
	return fHeight;
}
float CRow::TotalHeight(void) const
{
	CRow *temp;
	float size;
	
	for(temp=children,size=0;temp!=NULL;temp=temp->next) {
		size+=temp->Height();
		if(temp->bExpanded) size+=temp->TotalHeight();
	}
	return size;
}
CField* CRow::FindField(int iFieldIdentifier)
{
	CField *temp;
	
	for(temp=fields;temp!=NULL;temp=temp->next)
		if(temp->iFieldIdentifier==iFieldIdentifier)
			return temp;
	return NULL;
}
CRow* CRow::SuperItem(void)
{
	return owner;
}
void CRow::SetExpanded(bool bExpanded)
{
	this->bExpanded=bExpanded;
}
CRow* CRow::EachItemUnder(CRow *rootrow,bool oneLevel,CRow* (*func)(CRow*,void*),void *data)
{
	CRow *temp,*result;
	
	for(temp=rootrow;temp!=NULL;temp=temp->next) {
		result=func(temp,data);
		if(result!=NULL) return result;
		if((temp->children!=NULL)&&(!oneLevel)) result=EachItemUnder(temp->children,oneLevel,func,data);
		if(result!=NULL) return result;
	}
	return NULL;
}
void CRow::Update(void)
{
	// Do nothing
}
CRow* CRow::EachItemUnder(bool oneLevel,CRow* (*func)(CRow*,void*),void *data)
{
	return CRow::EachItemUnder(children,oneLevel,func,data);
}
float CRow::FindFontHeight(BFont *font)
{
	font_height fh;
	
	font->GetHeight(&fh);
	return fh.ascent+fh.descent+fh.leading+4;
}
bool CRow::IsSelected(void)
{
	return (selection_last!=NULL)||(selection_next!=NULL);
}
void CRow::Deselect(void)
{
	if(selection_last!=NULL) selection_last->selection_next=selection_next;
	if(selection_next!=NULL) selection_next->selection_last=selection_last;
	selection_last=selection_next=NULL;
}

// CColumn

CColumn::CColumn(float width,float minWidth,float maxWidth)
{
	fWidth=width;
	fMinWidth=minWidth;
	fMaxWidth=maxWidth;
	parent=NULL;
	bVisible=true;
}
CColumn::~CColumn()
{
	Unlink();
}
void CColumn::Unlink(void)
{
	if(parent!=NULL) {
		if(last!=NULL) last->next=next;
		if(next!=NULL) next->last=last;
		if(this==(*parent)) (*parent)=(next!=NULL)?next:last;
		parent=NULL;
	}
}
void CColumn::Attached(CColumn **parent,int iLevel)
{
	CColumn *temp,*lastfound=NULL;
	bool bDone;

	this->parent=parent;
	this->iLevel=iLevel;
	// Search linked list for place to insert, and do so.
	for(temp=(*parent),bDone=false;temp!=NULL;temp=temp->next) {
		if(temp->Level()>iLevel) {
			last=temp->last;
			if(last!=NULL) last->next=this;
			else (*parent)=this;
			next=temp;
			temp->last=this;
			bDone=true;
			break;
		}
		lastfound=temp;
	}
	if((*parent)==NULL) {
		last=next=NULL;
		(*parent)=this;
	} else if(!bDone) {
		lastfound->next=this;
		last=lastfound;
		next=NULL;
	}
}
float CColumn::Width(void) const
{
	return fWidth;
}
void CColumn::SetWidth(float width)
{
	if(width<fMinWidth) fWidth=fMinWidth;
	else if(width>fMaxWidth) fWidth=fMaxWidth;
	else fWidth=width;
}
float CColumn::MinWidth(void) const
{
	return fMinWidth;
}
float CColumn::MaxWidth(void) const
{
	return fMaxWidth;
}
void CColumn::DrawTitle(BView *targetView,BRect rect,seltype selected,bool bTransparent)
{
	// Draw the 3D box around the title, in case higher level code requires it
	float alpha=bTransparent?128:0;

	rgb_color colour=targetView->ViewColor();
	colour.alpha=(uint8)alpha;
	targetView->SetHighColor(colour);
	targetView->FillRect(rect);
	
	targetView->SetHighColor(128,128,128,alpha);
	targetView->StrokeRect(rect);
	
	targetView->SetHighColor(255,255,255,alpha);
	targetView->StrokeLine(BPoint(rect.left+1,rect.top+1),BPoint(rect.right-1,rect.top+1));
	targetView->StrokeLine(BPoint(rect.left+1,rect.top+1),BPoint(rect.left+1,rect.bottom-1));
	targetView->SetHighColor(192,192,192,alpha);
	targetView->StrokeLine(BPoint(rect.right-1,rect.top+1),BPoint(rect.right-1,rect.bottom-1));
	targetView->StrokeLine(BPoint(rect.left+1,rect.bottom-1),BPoint(rect.right-1,rect.bottom-1));
}
void CColumn::DrawField(BView *targetView,BRect rect,CField *rowdata,bool bSelected,rgb_color highlight)
{
	// Draw the background to assist the higher level code
	if(bSelected) {
		targetView->SetLowColor(highlight);
	} else {
		targetView->SetLowColor(targetView->ViewColor());
	}
	targetView->FillRect(rect,B_SOLID_LOW);
}
int CColumn::CompareFields(CField *row1,CField *row2)
{
	// I don't know how to compare nothing
	return 0;
}
bool CColumn::IsColumnVisible(void)
{
	return bVisible;
}
int CColumn::Level(void) const
{
	return iLevel;
}
CColumn* CColumn::FindColumn(CColumn *root,int iFieldIdentifier)
{
	CColumn *temp;
	
	for(temp=root;temp!=NULL;temp=temp->next)
		if(temp->iFieldIdentifier==iFieldIdentifier)
			return temp;
	return NULL;
}

// CColumnListView

CColumnListView::CColumnListView(BRect frame,char *name,uint32 resizingMode,uint32 drawFlags,uint32 listMode)
:BView(frame,name,resizingMode,drawFlags|B_WILL_DRAW|B_NAVIGABLE),
 BInvoker()
{
	BRect rect,bounds;
	const rgb_color defaulthighlight={192,192,255,0};

	bFastAdd=false;
	
	bAllowMultiple=listMode&LISTMODE_MULTI;
	
	bounds=Bounds();
	rows=NULL;
	columns=NULL;
	selection=NULL;

	rect.left=rect.top=0;
	rect.right=bounds.right;
	rect.bottom=rect.top+16.0;
	AddChild(headers=new CColumnHeaderView(this,rect,"headerview",B_FOLLOW_LEFT_RIGHT|B_FOLLOW_TOP));

	rect.top=rect.bottom+1.0;
	rect.bottom=bounds.bottom;
	rect.right=bounds.right-(B_V_SCROLL_BAR_WIDTH+1);
	AddChild(items=new CColumnItemView(this,rect,"bodyview",B_FOLLOW_ALL_SIDES));

	rect.top=17.0;
	rect.bottom=bounds.bottom;
	rect.right=bounds.right;
	rect.left=rect.right-B_V_SCROLL_BAR_WIDTH;
	AddChild(verticalScroll=new BScrollBar(rect,"vscroller",items,0,0,B_VERTICAL));
	
	bSortAscending=true;
	bSortEnable=false;
	
	highlight=defaulthighlight;
}
void CColumnListView::StartManyAdd(void)
{
	bFastAdd=true;
}
void CColumnListView::EndManyAdd(void)
{
	if(bFastAdd) {
		bFastAdd=false;
		items->Invalidate();
		items->RecalculateScroll();
	}
}

CColumnListView::~CColumnListView()
{
	while(rows!=NULL) delete rows;
	while(columns!=NULL) delete columns;
}
void CColumnListView::AddColumn(CColumn *column,int iFieldIdentifier,char *name,int iLevel)
{
	column->Attached(&columns,iLevel);
	column->iFieldIdentifier=iFieldIdentifier;
	// Do context menu
	if(name!=NULL) headers->DoContextMenu(column,name);
}
void CColumnListView::AddRow(CRow *row,CRow *parent)
{
	BRect bounds,item;
	float y,h;

	bounds=item=items->Bounds();
	h=row->Height();
	if(parent==NULL) {
		row->Attached(NULL,&rows);
		if(!bFastAdd) {
			item.OffsetBy(0,-h);
			items->CopyBits(item,bounds);
		}
	} else {
		parent->AddChild(row);
		if(!bFastAdd) {
			y=0;
			if(items->FindRowRect(rows,row,&y,&item)) {
				bounds.top=item.top;
				items->CopyBits(bounds,BRect(bounds.left,bounds.top+h,bounds.right,bounds.bottom+h));
			}
			items->Invalidate(row);
		}
	}
	if(!bFastAdd)
		items->RecalculateScroll();
}
void CColumnListView::DoSort(int iFieldIdentifier,bool bAscending)
{
	CListViewSorter sorter;
	CColumn *column=CColumn::FindColumn(columns,iFieldIdentifier);
	if(column!=NULL) {
		sorter.Sort(column,&rows,bAscending);
		items->Invalidate();
	}
}
bool CColumnListView::IsSorting(void) const
{
	return bSortEnable;
}
void CColumnListView::SetSorting(bool bSort)
{
	bSortEnable=bSort;
	if(IsSorting()) DoSort(GetSortColumnSelection(),IsSortAscending());
}
int CColumnListView::GetSortColumnSelection(void) const
{
	if(headers->selected!=NULL) return headers->selected->iFieldIdentifier;
	else return -1;
}
void CColumnListView::SetSortColumnSelection(int iFieldIdentifier)
{
	CColumn *column=CColumn::FindColumn(columns,iFieldIdentifier);
	if(column!=NULL) headers->SelectSort(column);
}
bool CColumnListView::IsSortAscending(void) const
{
	return bSortAscending;
}
void CColumnListView::SetSortAscending(bool bAscending)
{
	bSortAscending=bAscending;
	headers->Invalidate(headers->FindColumnRect(headers->selected));
	if(IsSorting()) DoSort(GetSortColumnSelection(),IsSortAscending());
}
void CColumnListView::ShowColumn(CColumn *column,bool bVisible)
{
	column->bVisible=bVisible;
	headers->Invalidate();
	items->Invalidate();
}
void CColumnListView::RemoveColumn(CColumn *column)
{
	column->Unlink();
	headers->Invalidate();
	items->Invalidate();
}
void CColumnListView::ExpandRow(CRow *row,bool bExpand)
{
	items->DoRowExpansion(row,bExpand);
}
void CColumnListView::RemoveRow(CRow *row)
{
	if(IsSelected(row)) {
		if(selection==row) selection=(row->selection_next!=NULL)?row->selection_next:row->selection_last;
		row->Deselect();
	}
	row->Unlink();
	headers->Invalidate();
	items->Invalidate();
	items->RecalculateScroll();
}
void CColumnListView::MakeEmpty(void)
{
	while(rows!=NULL) delete rows;
	selection=NULL;
	items->Invalidate();
	items->RecalculateScroll();
}
CRow* CColumnListView::SearchForNext(CRow *current,bool (*func)(CRow*,void*),void *data)
{
	do {
		current=FindNextItem(current);
		if((current!=NULL)&&func(current,data)) return current;
	} while(current!=NULL);
	return NULL;
}
CRow* CColumnListView::FindNextItem(CRow *row,bool bIgnoreChildren)
{
	if(row==NULL) return rows;
	if((row->children!=NULL)&&(!bIgnoreChildren)) return row->children;
	if(row->next!=NULL) return row->next;
	// Now we must check the linked list above us in the heirarchy
	if(row->owner!=NULL) return FindNextItem(row->owner,true);
	return NULL;
}
CRow* CColumnListView::CurrentSelection(int index)
{
	CRow *row;
	int count;
	
	for(row=selection,count=0;(row!=NULL)&&(count<index);row=row->selection_next,count++);
	return row;
}
void CColumnListView::SetCurrentSelection(CRow *newselection,bool bAdd)
{
	CRow *old;
	
	if((!bAllowMultiple)||(selection==NULL)) bAdd=false;
	if(bAdd) {
		if(IsSelected(newselection)) {
			if(selection==newselection) selection=(newselection->selection_next!=NULL)?newselection->selection_next:newselection->selection_next;
			newselection->Deselect();
		} else {
			if(selection!=NULL) selection->selection_last=newselection;
			newselection->selection_next=selection;
			selection=newselection;
		}
		Invalidate(newselection);
	} else {
		if(selection!=NULL) while(selection->selection_last!=NULL) selection=selection->selection_last;
		while(selection!=NULL) {
			old=selection;
			selection=old->selection_next;
			old->Deselect();
			Invalidate(old);
		}
		if(newselection!=NULL) {
			selection=newselection;
			Invalidate(selection);
			// Should I do this here?
			DoNotify(nfSelection);
		}
	}
}
void CColumnListView::ScrollTo(CRow *item)
{
	BRect rect,bounds;
	float y,f;
	CRow *row;
	
	if(item==NULL) return;
	// Make sure the list is sufficiently expanded to display
	for(row=item->owner;row!=NULL;row=row->owner)
		ExpandRow(row,true);
	// Scroll to it
	rect=bounds=items->Bounds();
	y=0;
	if(!items->FindRowRect(rows,item,&y,&rect)) return;
	if((rect.top<bounds.top)||(rect.bottom>bounds.bottom)) {
		if(rect.top<bounds.top)
			f=rect.top;
		else
			f=rect.top-(bounds.bottom-bounds.top)+item->Height();
		verticalScroll->SetValue(f);
	}
}
void CColumnListView::Invalidate(CRow *row)
{
	items->Invalidate(row);
}
void CColumnListView::Invalidate(CColumn *column)
{
	items->Invalidate(column);
}
void CColumnListView::Invalidate(CRow *row,int iFieldIdentifier)
{
	items->Invalidate(row,iFieldIdentifier);
}
void CColumnListView::SetHeaderBackground(rgb_color colour)
{
	headers->SetViewColor(colour);
	headers->Invalidate();
}
rgb_color CColumnListView::HeaderBackground(void) const
{
	return headers->ViewColor();
}
void CColumnListView::SetListBackground(rgb_color colour)
{
	items->SetViewColor(colour);
	items->Invalidate();
}
rgb_color CColumnListView::ListBackground(void) const
{
	return items->ViewColor();
}
void CColumnListView::SetHighlight(rgb_color colour)
{
	highlight=colour;
	if(selection!=NULL) items->Invalidate(selection);
}
rgb_color CColumnListView::Highlight(void) const
{
	return highlight;
}
void CColumnListView::DoNotify(NFACTION action)
{
	if(Message()!=NULL) {
		BMessage copy(*Message());
		copy.AddInt32("nfaction",action);
		InvokeNotify(&copy);
	}
}
void CColumnListView::Invalidate(void)
{
	headers->Invalidate();
	items->Invalidate();
	BView::Invalidate();
}
void CColumnListView::ContextMenuHook(BPoint point,uint32 buttons)
{
	// We need do nothing, at this point in time
}
CRow* CColumnListView::EachItemUnder(CRow *rootrow,bool oneLevel,CRow* (*func)(CRow*,void*),void *data)
{
	if(rootrow==NULL) return CRow::EachItemUnder(rows,oneLevel,func,data);
	else return CRow::EachItemUnder(rootrow->children,oneLevel,func,data);
}
void CColumnListView::ScrollToSelection(void)
{
	if(selection!=NULL) ScrollTo(selection);
}
void CColumnListView::MakeFocus(bool focused)
{
	items->MakeFocus(focused);
}
void CColumnListView::SetFont(const BFont *font,uint32 properties)
{
	items->SetFont(font,properties);
	BView::SetFont(font,properties);
}
BMessage* CColumnListView::DragRequested(CRow *row)
{
	return NULL;	// No drag in default implementation
}
CRow* CColumnListView::RowAt(int position,CRow *from,CRow *under)
{
	if(from==NULL) from=(under==NULL)?rows:under->children;
	while((position!=0)&&(from!=NULL)) {
		if(position>0) {
			position--;
			from=from->next;
		} else {
			position++;
			from=from->last;
		}
	}
	return from;
}
void CColumnListView::MessageReceived(BMessage *message)
{
	float delta;

	switch(message->what) {
		case B_MOUSE_WHEEL_CHANGED:
			{
				if((message->FindFloat("be:wheel_delta_y",&delta)==B_OK)&& (items->IsFocus()) /*valid*/ ) {
					verticalScroll->SetValue(verticalScroll->Value()+delta);
				}
			}
			break;
		default:
			BView::MessageReceived(message);
			break;
	}
}
void CColumnListView::AllAttached(void)
{
	BView::AllAttached();
}
bool CColumnListView::IsSelected(CRow *row)
{
	return row->IsSelected()||(selection==row);
}
void CColumnListView::Key_Up(void)
{
	char c=B_UP_ARROW;
	items->KeyDown(&c,1);
}
void CColumnListView::Key_Down(void)
{
	char c=B_DOWN_ARROW;
	items->KeyDown(&c,1);
}
void CColumnListView::Key_Left(void)
{
	char c=B_LEFT_ARROW;
	items->KeyDown(&c,1);
}
void CColumnListView::Key_Right(void)
{
	char c=B_RIGHT_ARROW;
	items->KeyDown(&c,1);
}

// CColumnHeaderView

CColumnHeaderView::CColumnHeaderView(CColumnListView *owner,BRect frame,char *name,uint32 resizingMode)
:BView(frame,name,resizingMode,B_WILL_DRAW)
{
	SetViewColor(216,216,216);
	this->owner=owner;
	selected=NULL;
	resizecolumn=NULL;
	menu=new BPopUpMenu("column_header_menu",false,false);
	menu->SetFont(be_plain_font);
	cursor=new BCursor(ColumnSizeCursor);
}
CColumnHeaderView::~CColumnHeaderView()
{
	delete menu;
	delete cursor;
}

void CColumnHeaderView::Draw(BRect updateRect)
{
	CColumn *temp;
	BRect colrect,myrect;

	myrect=Bounds();
	// Draw the pre-column lead-in
	colrect=myrect;
	colrect.left=-2;
	colrect.right=COLLEADIN-COLSEP;
	if(updateRect.Intersects(colrect)) {
		SetHighColor(128,128,128);
		StrokeRect(colrect);
		SetHighColor(255,255,255);
		StrokeLine(BPoint(colrect.left+1,colrect.top+1),BPoint(colrect.right-1,colrect.top+1));
		SetHighColor(192,192,192);
		StrokeLine(BPoint(colrect.right-1,colrect.top+1),BPoint(colrect.right-1,colrect.bottom-1));
		StrokeLine(BPoint(colrect.left+1,colrect.bottom-1),BPoint(colrect.right-1,colrect.bottom-1));
	}
	// Draw the actual columns
	colrect=myrect;
	colrect.left=COLLEADIN;
	for(temp=owner->columns;temp!=NULL;temp=temp->next) {
		if(temp->bVisible) {
			colrect.right=colrect.left+temp->fWidth;
			if(updateRect.Intersects(colrect)) {
				PushState();
				temp->DrawTitle(this,colrect,(selected==temp)?(owner->IsSortAscending()?CColumn::slAscending:CColumn::slDescending):CColumn::slNone);
				PopState();
			}
			colrect.left=colrect.right+COLSEP;
		}
	}
	// Draw the post-column lead-out
	colrect.right=myrect.right+2;
	if(updateRect.Intersects(colrect)) {
		SetHighColor(128,128,128);
		StrokeRect(colrect);
		SetHighColor(255,255,255);
		StrokeLine(BPoint(colrect.left+1,colrect.top+1),BPoint(colrect.right-1,colrect.top+1));
		StrokeLine(BPoint(colrect.left+1,colrect.top+1),BPoint(colrect.left+1,colrect.bottom-1));
		SetHighColor(192,192,192);
		StrokeLine(BPoint(colrect.left+1,colrect.bottom-1),BPoint(colrect.right-1,colrect.bottom-1));
	}
}

CColumn* CColumnHeaderView::FindColumn(BPoint point)
{
	CColumn *temp;
	BRect colrect;
	
	colrect=Bounds();
	colrect.left=16;
	for(temp=owner->columns;temp!=NULL;temp=temp->next) {
		if(temp->bVisible) {
			colrect.right=colrect.left+temp->fWidth;
			if(colrect.Contains(point))
				return temp;
			colrect.left=colrect.right+COLSEP;
		}
	}
	return NULL;
}
CColumn* CColumnHeaderView::FindResizeColumn(BPoint point)
{
	CColumn *temp;
	BRect colrect,dragrect;
	
	colrect=Bounds();
	dragrect=colrect;
	colrect.left=16;
	for(temp=owner->columns;temp!=NULL;temp=temp->next) {
		if(temp->bVisible) {
			colrect.right=colrect.left+temp->fWidth;
			dragrect.left=colrect.right-1;
			dragrect.right=colrect.right+3;
			if(dragrect.Contains(point)) return temp;
			colrect.left=colrect.right+COLSEP;
		}
	}
	return NULL;
}
BRect CColumnHeaderView::FindColumnRect(CColumn *column)
{
	CColumn *temp;
	BRect colrect;
	
	colrect=Bounds();
	colrect.left=16;
	for(temp=owner->columns;temp!=NULL;temp=temp->next) {
		if(temp->bVisible) {
			colrect.right=colrect.left+temp->fWidth;
			if(column==temp) return colrect;
			colrect.left=colrect.right+2;
		}
	}
	return colrect;
}

BBitmap* CColumnHeaderView::MakeColumnBitmap(CColumn *column)
{
	BBitmap *bitmap,*donebitmap;
	BRect rect;
	BView *view;
	
	rect.left=rect.top=0;
	rect.right=column->fWidth;
	rect.bottom=16;
	bitmap=new BBitmap(rect,B_RGB32,true);
	bitmap->Lock();
	bitmap->AddChild(view=new BView(rect,"bitmap_view",B_FOLLOW_ALL_SIDES,0));
	column->DrawTitle(view,rect,CColumn::slNone,true);
	view->Sync();
	bitmap->Unlock();
	donebitmap=new BBitmap(rect,B_RGB32);
	memcpy(donebitmap->Bits(),bitmap->Bits(),bitmap->BitsLength());
	delete bitmap;
	return donebitmap;
}

void CColumnHeaderView::MouseDown(BPoint point)
{
	uint32 buttons;
	BPoint newpoint;
	float f,g;
	BMessage *message;
	CColumn *column;
	BRect rect;
	BBitmap *bitmap;
	CColumnMenuItem *menuitem;
	int i,j;
	
	GetMouse(&newpoint,&buttons,false);
	if(buttons&B_SECONDARY_MOUSE_BUTTON) {
		// Discover location for context menu
		ConvertToScreen(&newpoint);
		newpoint.x-=1;
		newpoint.y-=1;
		// Mark/unmark menu options
		j=menu->CountItems();
		for(i=0;i<j;i++) {
			menuitem=(CColumnMenuItem*)menu->ItemAt(i);
			menuitem->SetMarked(menuitem->column->IsColumnVisible());
		}
		// Actually do context menu
		menuitem=(CColumnMenuItem*)menu->Go(newpoint,false,true,false);
		// Check
		if(menuitem!=NULL) {
			menuitem->column->bVisible=!menuitem->column->bVisible;
			Invalidate();
			owner->items->Invalidate();
			owner->DoNotify(CColumnListView::nfResize);
		}
	} else if(buttons&B_PRIMARY_MOUSE_BUTTON) {
		column=FindResizeColumn(point);
		if(column!=NULL) {
			// The column is being resized
			resizecolumn=column;
			resizepos=newpoint.x;
			SetMouseEventMask(B_POINTER_EVENTS,B_LOCK_WINDOW_FOCUS|B_NO_POINTER_HISTORY);
			return;
		}
		column=FindColumn(point);
		rect=FindColumnRect(column);
		if(column==NULL) return;
		while(buttons&B_PRIMARY_MOUSE_BUTTON) {
			GetMouse(&newpoint,&buttons,false);
			f=newpoint.x-point.x;
			g=newpoint.y-point.y;
			if(f<0) f=-f;
			if(g<0) g=-g;
			if((f>DRAGTHRESHOLD)||(g>DRAGTHRESHOLD)) {
				message=new BMessage(COLUMNMOVE);
				message->AddPointer("header",this);
				message->AddPointer("column",column);
				bitmap=MakeColumnBitmap(column);
				DragMessage(message,bitmap,B_OP_ALPHA,BPoint(point.x-rect.left,point.y),this);
				delete message;
				return;
			}
		}
		// If we reach here, the user wants to select this group
		SelectSort(column);
	}
	BView::MouseDown(point);
}
void CColumnHeaderView::SelectSort(CColumn *column)
{
	CColumn *old;

	if(column==selected) owner->SetSortAscending(!owner->IsSortAscending());
	else {
		old=selected;
		selected=column;
		if(old!=NULL) Invalidate(FindColumnRect(old));
		if(selected!=NULL) Invalidate(FindColumnRect(selected));
	}
	if(owner->IsSorting()) owner->DoSort(column->iFieldIdentifier,owner->IsSortAscending());
	owner->DoNotify(CColumnListView::nfSortOption);
}
void CColumnHeaderView::FixColumnLevels(int level,int orig)
{
	CColumn *temp;
	
	if(level>orig) {
		for(temp=owner->columns;temp!=NULL;temp=temp->next)
			if((temp->iLevel<=level)&&(temp->iLevel>orig)) temp->iLevel--;
	} else {
		for(temp=owner->columns;temp!=NULL;temp=temp->next)
			if((temp->iLevel>=level)&&(temp->iLevel<orig)) temp->iLevel++;
	}
}
void CColumnHeaderView::MouseMoved(BPoint point,uint32 transit,const BMessage *message)
{
	BRect rect,rect2,colrect;
	CColumnHeaderView *that;
	CColumn *column,*cur;
	int level;
	float f,x1,x2;
	BCursor *usecursor=(BCursor*)B_CURSOR_SYSTEM_DEFAULT;

	if(resizecolumn!=NULL) {
		colrect=FindColumnRect(resizecolumn);
		rect2=owner->items->Bounds();
		rect=Bounds();
		f=resizecolumn->Width();
		rect2.left=colrect.left+f;
		resizecolumn->SetWidth(f-resizepos+point.x);
		resizepos=point.x;
		rect.left=colrect.left;
		if(rect2.left>(colrect.left+resizecolumn->Width()))
			rect2.left=colrect.left+resizecolumn->Width();
		ColumnResized(colrect,resizecolumn,f);
		usecursor=cursor;
	} else if(FindResizeColumn(point)!=NULL) usecursor=cursor;
	be_app->SetCursor(usecursor);
	if((message!=NULL)&&(message->what==COLUMNMOVE)) {
		message->FindPointer("header",(void**)&that);
		if(that==this) {
			message->FindPointer("column",(void**)&column);
			// We have a column drag in operation, so we do something
			switch(transit) {
				case B_EXITED_VIEW:
					// Assume the column is going invisible
					column->bVisible=false;
					Invalidate();
					owner->items->Invalidate();
					break;
				case B_OUTSIDE_VIEW:
					// Not much to do, assume the user doesn't want the column to
					// be visible any more.
					break;
				case B_ENTERED_VIEW:
					// Assume the column is going visible
					column->bVisible=true;
					Invalidate();
					owner->items->Invalidate();
					// Fall through to process the possible positions it occupies
				case B_INSIDE_VIEW:
					// We must use our incredible detective skills to detect
					// where the column is being dragged, and move it there.
					for(cur=owner->columns,level=0,f=16;cur!=NULL;cur=cur->next,level++) {
						if((column!=cur)&&(column->next!=cur)) {
							if(cur->last==NULL) x1=0;
							else x1=f-(cur->last->fWidth/4);
							x2=f+(cur->fWidth/4);
							if((point.x>=x1)&&(point.x<=x2)) {
								column->Unlink();
								FixColumnLevels(level,column->iLevel);
								column->Attached(&(owner->columns),level);
								Invalidate();
								owner->items->Invalidate();
								break;
							}
						}
						f+=cur->fWidth;
					}
					if(point.x>f) {
						column->Unlink();
						FixColumnLevels(level,column->iLevel);
						column->Attached(&(owner->columns),level);
						Invalidate();
						owner->items->Invalidate();
					}
					break;
			}
		}
	}
	BView::MouseMoved(point,transit,message);
}
void CColumnHeaderView::MouseUp(BPoint point)
{
	if(resizecolumn!=NULL) {
		resizecolumn=NULL;
		owner->DoNotify(CColumnListView::nfResize);
		be_app->SetCursor(B_CURSOR_SYSTEM_DEFAULT);
	}
	BView::MouseUp(point);
}
void CColumnHeaderView::MessageReceived(BMessage *message)
{
	switch(message->what) {
		case COLUMNMOVE:
			owner->DoNotify(CColumnListView::nfResize);
			break;
		default:
			BView::MessageReceived(message);
			break;
	}
}
void CColumnHeaderView::DoContextMenu(CColumn *column,char *name)
{
	CColumnMenuItem *item;
	
	item=new CColumnMenuItem(column,name,column->iFieldIdentifier);
	menu->AddItem(item);
}
void CColumnHeaderView::ColumnResized(BRect columnrect,CColumn *column,float oldwidth)
{
	float newwidth=column->Width();
	BRect invalidrect,copyfrom,copyto;
	
	if(newwidth==oldwidth) return;
	invalidrect=copyfrom=copyto=Bounds();
	copyfrom.left=columnrect.left+oldwidth-(column->iLevel*3);
	copyto.left=columnrect.left+newwidth-(column->iLevel*3);
	copyto.right=copyto.left+(copyfrom.right-copyfrom.left);
	if(newwidth<oldwidth) {	// Resized left
		invalidrect.left=copyto.right;
	} else {				// Resized right
		invalidrect.left=columnrect.left+oldwidth-(column->iLevel*3)-2;
		invalidrect.right=copyto.left;
	}
	CopyBits(copyfrom,copyto);
	Invalidate(invalidrect);
	owner->items->ColumnResized(columnrect,column,oldwidth);
	owner->DoNotify(CColumnListView::nfResize);
}

// CColumnItemView

CColumnItemView::CColumnItemView(CColumnListView *owner,BRect frame,char *name,uint32 resizingMode)
:BView(frame,name,resizingMode,B_WILL_DRAW|B_FRAME_EVENTS|B_NAVIGABLE)
{
	SetViewColor(255,255,255);
	this->owner=owner;
}
CColumnItemView::~CColumnItemView()
{
}
void CColumnItemView::Draw(BRect updateRect)
{
	CColumn *column;
	float y;
	
	iLevel=0;
	for(column=owner->columns;column!=NULL;column=column->next) {
		if((iLevel==column->iLevel)&&(!column->IsColumnVisible())) {
			iLevel++;
			column=owner->columns;
		}
	}
	y=0;
	bDrawDone=false;
	DoDrawLoop(updateRect,owner->rows,0,&y);
	Sync();		// For any icons
	ConstrainClippingRegion(NULL);
}
void CColumnItemView::DoDrawLoop(BRect updateRect,CRow *rowroot,float x,float *y)
{
	CRow *row;
	CColumn *column;
	CField *field;
	BRect rect,linerect,bounds;
	BRegion region;
	float h,w;
	bool selected;

	bounds=Bounds();
	
	for(row=rowroot;row!=NULL;row=row->next) {
		h=row->Height();
		rect=bounds;
		rect.top=(*y);
		rect.bottom=(*y)+h;
		if(rect.Intersects(updateRect)) {
			if(!bDrawDone) {
				bDrawDone=true;
			}
		} else {
			if(bDrawDone) return;
			else {
				(*y)+=h;
				if(row->IsExpanded()) DoDrawLoop(updateRect,row->children,x+16,y);
				continue;
			}
		}

		row->Update();

		selected=owner->IsSelected(row);

		if(selected) {
			rect.top=(*y);
			rect.bottom=(*y)+h-1;
			rect.left=x+EXPANDERWIDTH;
			rect.right=bounds.right;
			SetHighColor(owner->highlight);
			FillRect(rect);
		}
		
		for(field=row->fields;field!=NULL;field=field->next) {
			column=CColumn::FindColumn(owner->columns,field->iFieldIdentifier);
			if((column!=NULL)&&column->IsColumnVisible()) {
				rect=owner->headers->FindColumnRect(column);
				rect.top=(*y);
				rect.bottom=(*y)+h-1;
				rect.right-=4;
				if(column->iLevel==iLevel) {
					linerect=rect;
					linerect.bottom+=1;	// Proper-square expander thingies
					linerect.left=x;
					linerect.right=x+EXPANDERWIDTH;
					region.Set(linerect);
					PushState();
					ConstrainClippingRegion(&region);
					row->DrawExpander(this,linerect);
					PopState();
					rect.left=linerect.right;
					// Do icon stuff
					linerect.left=rect.left;
					linerect.right=bounds.right;
					region.Set(linerect);
					PushState();
					ConstrainClippingRegion(&region);
					if(row->DrawIcon(this,linerect,&w)) {
						rect.left+=w;
					}
					PopState();
				}
				if(rect.Intersects(updateRect)) {
					region.Set(rect);
					PushState();
					ConstrainClippingRegion(&region);
					column->DrawField(this,rect,field,selected,owner->highlight);
					PopState();
				}
			}
		}
		(*y)+=h;
		if(row->IsExpanded()) DoDrawLoop(updateRect,row->children,x+16,y);
	}
}
void CColumnItemView::MouseDown(BPoint point)
{
	BPoint temp;
	CRow *row,*rowexp;
	float x,y,f,g;
	uint32 buttons,tempbuttons;
	BRect rect;
	
	MakeFocus();
	y=0;
	row=FindRow(point,owner->rows,0,&y);
	GetMouse(&temp,&buttons,false);
	tempbuttons=buttons;
	while((buttons&B_PRIMARY_MOUSE_BUTTON)&&(owner->selection==row)) {
		GetMouse(&temp,&buttons,false);
		f=temp.x-point.x;
		g=temp.y-point.y;
		if(f<0) f=-f;
		if(g<0) g=-g;
		if((f>DRAGTHRESHOLD)||(g>DRAGTHRESHOLD)) {
			BMessage *dragmessage;
			dragmessage=owner->DragRequested(row);
			if(dragmessage!=NULL) {
				y=0;
				rect=Bounds();
				FindRowRect(owner->rows,row,&y,&rect);
				DragMessage(dragmessage,rect,owner);
				delete dragmessage;
				return;
			}
		}
	}
	buttons=tempbuttons;
	x=y=0;
	rowexp=FindExpander(point,owner->rows,0,&y,&x);
	if(rowexp!=NULL) DoRowExpansion(rowexp,!rowexp->IsExpanded());
	else if(buttons&B_PRIMARY_MOUSE_BUTTON) {
		// Item selection
		if(row!=NULL) owner->SetCurrentSelection(row,MULTISELECT());
	}
	ConvertToScreen(&point);
	owner->ConvertFromScreen(&point);
	owner->ContextMenuHook(point,buttons);
}
CRow* CColumnItemView::FindRow(BPoint point,CRow *rowroot,float x,float *y)
{
	CRow *row,*temprow;
	float h;
	BRect rect;

	rect=Bounds();
	rect.left=x+EXPANDERWIDTH;
	for(row=rowroot;row!=NULL;row=row->next) {
		h=row->Height();
		rect.top=(*y);
		rect.bottom=(*y)+h;
		if(rect.Contains(point)) {
			return row;
		}
		(*y)+=h;
		if(row->IsExpanded()) {
			temprow=FindRow(point,row->children,x+16,y);
			if(temprow!=NULL) return temprow;
		}
	}
	return NULL;
}
CRow* CColumnItemView::FindExpander(BPoint point,CRow *rowroot,float x,float *y,float *fx)
{
	CRow *row,*temprow;
	CColumn *column;
	CField *field;
	BRect rect;
	float h;
	int iLevel;
	
	iLevel=0;
	for(column=owner->columns;column!=NULL;) {
		if((iLevel==column->iLevel)&&(!column->IsColumnVisible())) {
			iLevel++;
			column=owner->columns;
		} else column=column->next;
	}
	for(row=rowroot;row!=NULL;row=row->next) {
		h=row->Height();
		for(field=row->fields;field!=NULL;field=field->next) {
			rect.top=(*y);
			rect.bottom=(*y)+h;
			rect.left=x;
			rect.right=x+EXPANDERWIDTH;
			if(row->IsPointWithinExpander(rect,point)) {
				(*fx)=x;
				return row;
			}
		}
		(*y)+=h;
		if(row->IsExpanded()) {
			temprow=FindExpander(point,row->children,x+16,y,fx);
			if(temprow!=NULL) return temprow;
		}
	}
	return NULL;
}
void CColumnItemView::ColumnResized(BRect columnrect,CColumn *column,float oldwidth)
{
	float newwidth=column->Width();
	BRect invalidrect,copyfrom,copyto;
	
	if(newwidth==oldwidth) return;
	invalidrect=copyfrom=copyto=Bounds();
	copyfrom.left=columnrect.left+oldwidth-(column->iLevel*3);
	copyto.left=columnrect.left+newwidth-(column->iLevel*3);
	copyto.right=copyto.left+(copyfrom.right-copyfrom.left);
	if(newwidth<oldwidth) {	// Resized left
		invalidrect.left=copyto.right;
	} else {				// Resized right
		invalidrect.left=columnrect.left+oldwidth-(column->iLevel*3)-4;
		invalidrect.right=copyto.left+2;
	}
	CopyBits(copyfrom,copyto);
	Invalidate(invalidrect);
}
void CColumnItemView::DoRowExpansion(CRow *row,bool bNewExpanded)
{
	BRect rect,rect2,rect3,bounds;
	float h,x,y;

	if(bNewExpanded==row->IsExpanded()) return;
	x=y=0;
	if(!GetExpanderXY(owner->rows,row,0,&x,&y)) return;
	row->ExpandOperation(bNewExpanded);
	h=row->TotalHeight();
	rect=rect2=rect3=bounds=Bounds();
	if(row->bExpanded) {	// Shunt contents down by row child height, and invalidate shunted area
		rect.top=y+row->Height();
		rect2.top=rect.top+h;
		rect2.bottom=rect2.top+(rect.bottom-rect.top);
		rect3.top=rect.top;
		rect3.bottom=rect3.top+h;
	} else {				// Shunt contents back up, and invalidate bottom
		rect2.top=y+row->Height();
		rect.top=rect2.top+h;
		rect.bottom=rect.top+(rect2.bottom-rect2.top);
		rect3.top=rect.top;
		rect3.bottom=bounds.bottom;
	}
	CopyBits(rect,rect2);
	Invalidate(rect3);
	Invalidate(BRect(x,y,x+EXPANDERWIDTH,y+row->Height()));
	RecalculateScroll();
	row->ExpandOperationComplete(bNewExpanded);
}
bool CColumnItemView::GetExpanderXY(CRow *rowroot,CRow *findrow,float x,float *fx,float *y)
{
	CRow *row;
	CColumn *column;
	BRect rect;
	float h;
	int iLevel;
	
	iLevel=0;
	for(column=owner->columns;column!=NULL;column=column->next)
		if((iLevel==column->iLevel)&&(!column->IsColumnVisible())) {
			iLevel++;
			column=owner->columns;
		}
	for(row=rowroot;row!=NULL;row=row->next) {
		h=row->Height();
		if(row==findrow) {
			rect.top=(*y);
			rect.bottom=(*y)+h;
			rect.left=x;
			rect.right=x+EXPANDERWIDTH;
			(*fx)=x;
			return true;
		}
		(*y)+=h;
		if(row->IsExpanded())
			if(GetExpanderXY(row->children,findrow,x+16,fx,y))
				return true;
	}
	return false;
}
void CColumnItemView::FrameResized(float width,float height)
{
	RecalculateScroll();
	BView::FrameResized(width,height);
}
void CColumnItemView::RecalculateScroll(void)
{
	float height,f;
	CRow *row;
	BRect rect;
	
	for(row=owner->rows,height=0;row!=NULL;row=row->next) {
		height+=row->Height();
		if(row->IsExpanded()) height+=row->TotalHeight();
	}
	rect=Bounds();
	f=height-(rect.bottom-rect.top);
	owner->verticalScroll->SetRange(0,(f<0)?0:f);
	owner->verticalScroll->SetSteps(16.0,rect.bottom-rect.top);
}
void CColumnItemView::Invalidate(CRow *row)
{
	float y=0;
	BRect rect=Bounds();
	if(!FindRowRect(owner->rows,row,&y,&rect)) return;
	Invalidate(rect);
}
void CColumnItemView::Invalidate(CColumn *column)
{
	if(!column->IsColumnVisible()) return;
	BRect rect=owner->headers->FindColumnRect(column),bounds=Bounds();
	rect.top=bounds.top;
	rect.bottom=bounds.bottom;
	Invalidate(rect);
}
void CColumnItemView::Invalidate(CRow *row,int iFieldIdentifier)
{
	float y;
	BRect rowrect,columnrect,fieldrect;
	CColumn *column;

	// Find column
	column=CColumn::FindColumn(owner->columns,iFieldIdentifier);
	if((column==NULL)||(!column->IsColumnVisible())) return;
	// Find row rectangle
	y=0;
	rowrect=Bounds();
	if(!FindRowRect(owner->rows,row,&y,&rowrect)) return;
	// Find column rectangle
	columnrect=owner->headers->FindColumnRect(column);
	// Find field rectangle
	fieldrect.left=columnrect.left;
	fieldrect.right=columnrect.right;
	fieldrect.top=rowrect.top;
	fieldrect.bottom=rowrect.bottom;
	// Invalidate the field
	Invalidate(fieldrect);
}
// When calling this function, ensure output is set to Bounds()
bool CColumnItemView::FindRowRect(CRow *rowroot,CRow *findrow,float *y,BRect *output)
{
	CRow *row;
	float h;
	
	for(row=rowroot;row!=NULL;row=row->next) {
		h=row->Height();
		if(row==findrow) {
			output->top=(*y);
			output->bottom=(*y)+h;
			return true;
		}
		(*y)+=h;
		if(row->IsExpanded())
			if(FindRowRect(row->children,findrow,y,output))
				return true;
	}
	return false;
}
inline void CColumnItemView::Invalidate(void)
{
	BView::Invalidate();
}
inline void CColumnItemView::Invalidate(BRect rect)
{
	BView::Invalidate(rect);
}
void CColumnItemView::KeyDown(const char *bytes,int32 numBytes)
{
	CRow *newselection;

	newselection=NULL;
	if(numBytes==1) {
		switch(bytes[0]) {
			case B_UP_ARROW:	// Move up to previous item
				if(owner->selection==NULL) newselection=owner->rows;
				else if(owner->selection->last!=NULL) {
					newselection=owner->selection->last;
					while((newselection!=NULL)&&newselection->IsExpanded()&&(newselection->children!=NULL)) {
						newselection=newselection->children;
						while(newselection->next!=NULL) newselection=newselection->next;
					}
				} else if(owner->selection->owner!=NULL) newselection=owner->selection->owner;
				break;
			case B_DOWN_ARROW:	// Move down to next item
				if(owner->selection==NULL) newselection=owner->rows;
				else if((owner->selection->children!=NULL)&&owner->selection->IsExpanded()) newselection=owner->selection->children;
				else if(owner->selection->next!=NULL) newselection=owner->selection->next;
				else if(owner->selection->owner!=NULL) {
					newselection=owner->selection;
					while((newselection!=NULL)&&(newselection->next==NULL)) newselection=newselection->owner;
					if(newselection!=NULL) newselection=newselection->next;
				}
				break;
			case B_LEFT_ARROW:	// Collapse selected items children
				if(owner->selection==NULL) newselection=owner->rows;
				else DoRowExpansion(owner->selection,false);
				break;
			case B_RIGHT_ARROW:	// Expand selected items children
				if(owner->selection==NULL) newselection=owner->rows;
				else DoRowExpansion(owner->selection,true);
				break;
			default:
				BView::KeyDown(bytes,numBytes);
				break;
		}
	} else BView::KeyDown(bytes,numBytes);
	if(newselection!=NULL) {
		owner->SetCurrentSelection(newselection,MULTISELECT());
		owner->ScrollTo(newselection);
	}
}
