#include <stdio.h>
#include <string.h>
#include <Bitmap.h>
#include "ListViewerTypes.h"

// CTitledColumn, for displaying a column with a title

// Constructor and destructor
CTitledColumn::CTitledColumn(const char *title,float width,float minWidth,float maxWidth)
:CColumn(width,minWidth,maxWidth)
{
	int x;
	
	x=strlen(title);
	this->title=new char[x+10];
	memcpy(this->title,title,x);
	this->title[x]='\0';
}
CTitledColumn::~CTitledColumn()
{
	delete[] title;
}

// Altering and returning the string
void CTitledColumn::SetString(char *title)
{
	int x;

	delete[] this->title;
	x=strlen(title);
	this->title=new char[x+10];
	memcpy(this->title,title,x);
	this->title[x]='\0';
}
const char* CTitledColumn::GetString(void)
{
	return title;
}

// Drawing the column title
void CTitledColumn::DrawTitle(BView *targetView,BRect rect,seltype selected,bool bTransparent)
{
	float alpha=bTransparent?128:0;

	CColumn::DrawTitle(targetView,rect,selected,bTransparent);
	targetView->SetHighColor(0,0,0,alpha);
	targetView->SetLowColor(216,216,216,alpha);
	targetView->DrawString(title,BPoint(rect.left+5,rect.bottom-4));
	if(selected!=slNone)
		targetView->StrokeLine(BPoint(rect.left+5,rect.bottom-2),
			BPoint(rect.left+5+be_plain_font->StringWidth(title),rect.bottom-2));
}

// CBitmapColumn, for displaying a column with a bitmap title

// Constructor and destructor
CBitmapColumn::CBitmapColumn(BBitmap *title,float width,float minWidth,float maxWidth)
:CColumn(width,minWidth,maxWidth)
{
	this->title=new BBitmap(title->Bounds(),title->ColorSpace());
	memcpy(this->title->Bits(),title->Bits(),title->BitsLength());
}
CBitmapColumn::~CBitmapColumn()
{
	delete title;
}

// For setting and returning the bitmap
void CBitmapColumn::SetBitmap(BBitmap *title)
{
	delete this->title;
	this->title=new BBitmap(title->Bounds(),title->ColorSpace());
	memcpy(this->title->Bits(),title->Bits(),title->BitsLength());
}
BBitmap* CBitmapColumn::GetBitmap(void)
{
	return title;
}

// Drawing the column title
void CBitmapColumn::DrawTitle(BView *targetView,BRect rect,seltype selected,bool bTransparent)
{
	BRect bumprect;

	CColumn::DrawTitle(targetView,rect,selected,bTransparent);
	targetView->SetDrawingMode(B_OP_ALPHA);
	bumprect=title->Bounds();
	targetView->DrawBitmapAsync(title,BPoint(rect.left+(((rect.right-rect.left)-(bumprect.right-bumprect.left))/2),rect.top+(((rect.bottom-rect.top)-(bumprect.bottom-bumprect.top))/2)));
}

// CStringField, for containing information about a single field in the view
CStringField::CStringField(char *string)
:CField()
{
	int x;

	x=strlen(string);
	this->string=new char[x+1];
	memcpy(this->string,string,x);
	this->string[x]='\0';
	colour.red=0;
	colour.green=0;
	colour.blue=0;
	colour.alpha=0;
	bBold=false;
}
CStringField::~CStringField()
{
	delete[] string;
}
void CStringField::SetString(char *string)
{
	int x;

	delete[] this->string;
	x=strlen(string);
	this->string=new char[x+1];
	memcpy(this->string,string,x);
	this->string[x]='\0';
}
char* CStringField::String(void) const
{
	return string;
}
void CStringField::SetColour(rgb_color colour)
{
	this->colour=colour;
}
rgb_color CStringField::Colour(void) const
{
	return colour;
}

bool CStringField::Bold(void)
{
	return bBold;
}
void CStringField::SetBold(bool bBold)
{
	this->bBold=bBold;
}

// CStringColumn, for drawing a CStringField onto the list view
CStringColumn::CStringColumn(const char *title,float width,float minWidth,float maxWidth)
:CTitledColumn(title,width,minWidth,maxWidth)
{
}
void CStringColumn::DrawField(BView *targetView,BRect rect,CField *rowdata,bool bSelected,rgb_color highlight)
{
	CStringField *field=(CStringField*)rowdata;
	BFont font,oaf;
	font_height fh;
	uint16 of;
	
	CTitledColumn::DrawField(targetView,rect,rowdata,bSelected,highlight);
	targetView->SetHighColor(field->colour);
	targetView->GetFont(&font);
	oaf=font;
	of=font.Face();
	if(field->Bold()) {
		font.SetFace(B_BOLD_FACE);
		font.SetSize(font.Size()+1);
	}
	font.SetFace(field->Bold()?B_BOLD_FACE:of);
	if(of!=font.Face()) targetView->SetFont(&font);
	font.GetHeight(&fh);
	targetView->DrawString(field->string,BPoint(rect.left+5,rect.bottom-(fh.descent+fh.leading)));
	targetView->SetFont(&oaf);
}
int CStringColumn::CompareFields(CField *field1,CField *field2)
{
	return strcmp(((CStringField*)field1)->string,((CStringField*)field2)->string);
}

// CIntField, for containing information on a single integer field in the view
CIntField::CIntField(int i)
:CField()
{
	this->i=i;
}
void CIntField::SetInteger(int i)
{
	this->i=i;
}
int CIntField::Integer(void) const
{
	return i;
}

// CIntColumn, for drawing a CIntField onto the list view
CIntColumn::CIntColumn(const char *title,float width,float minWidth,float maxWidth)
:CTitledColumn(title,width,minWidth,maxWidth)
{
}
void CIntColumn::DrawField(BView *targetView,BRect rect,CField *rowdata,bool bSelected,rgb_color highlight)
{
	char buf[20];
	BFont font;
	font_height fh;

	CTitledColumn::DrawField(targetView,rect,rowdata,bSelected,highlight);
	targetView->SetHighColor(0,0,0);
	sprintf(buf,"%i",((CIntField*)rowdata)->i);
	targetView->GetFont(&font);
	font.GetHeight(&fh);
	targetView->DrawString(buf,BPoint(rect.left+5,rect.bottom-(fh.descent+fh.leading)));
}
int CIntColumn::CompareFields(CField *row1,CField *row2)
{
	return ((CIntField*)row1)->i-((CIntField*)row2)->i;
}

// CDateField, for storing a date
CDateField::CDateField(uint32 time)
:CField()
{
	fieldtime=time;
	string=NULL;
}
CDateField::~CDateField()
{
	if(string!=NULL) delete[] string;
}
void CDateField::SetTime(uint32 time)
{
	fieldtime=time;
	if(string!=NULL) delete[] string;
	string=MakeDateString(fieldtime);
}
uint32 CDateField::Time(void) const
{
	return fieldtime;
}
char* CDateField::MakeDateString(time_t time)
{
	struct tm *bob;
	char *months[]={"January","February","March","April","May","June","July","August","September","October","November","December"};
	char *skrill,*s;
	int hr;
	
	bob=localtime(&time);
	skrill=new char[100];
	hr=bob->tm_hour;
	if(hr>=12) {
		if(hr>12) hr-=12;
		s="PM";
	} else s="AM";
	sprintf(skrill,"%i:%s%i:%s%i%s - %i %s %i",
		hr,(bob->tm_min<10)?"0":"",bob->tm_min,
		(bob->tm_sec<10)?"0":"",bob->tm_sec,s,
		bob->tm_mday,months[bob->tm_mon],bob->tm_year+1900);
	return skrill;
}

// CDateColumn, for drawing CDateFields onto the list view
CDateColumn::CDateColumn(const char *title,float width,float minWidth,float maxWidth)
:CTitledColumn(title,width,minWidth,maxWidth)
{
}
void CDateColumn::DrawField(BView *targetView,BRect rect,CField *rowdata,bool bSelected,rgb_color highlight)
{
	BFont font;
	font_height fh;

	if(((CDateField*)rowdata)->string==NULL) ((CDateField*)rowdata)->string=((CDateField*)rowdata)->MakeDateString(((CDateField*)rowdata)->fieldtime);
	CTitledColumn::DrawField(targetView,rect,rowdata,bSelected,highlight);
	targetView->SetHighColor(0,0,0);
	targetView->GetFont(&font);
	font.GetHeight(&fh);
	targetView->DrawString(((CDateField*)rowdata)->string,BPoint(rect.left+5,rect.bottom-(fh.descent+fh.leading)));
}
int CDateColumn::CompareFields(CField *row1,CField *row2)
{
	return ((CDateField*)row1)->fieldtime-((CDateField*)row2)->fieldtime;
}

// Bitmap stuff
CBitmapField::CBitmapField(BBitmap *title,int level,bool bVisible)
:CField()
{
	bitmap=new BBitmap(title->Bounds(),title->ColorSpace());
	memcpy(bitmap->Bits(),title->Bits(),title->BitsLength());
	value=level;
	this->bVisible=bVisible;
}
CBitmapField::~CBitmapField()
{
	delete bitmap;
}
void CBitmapField::SetBitmap(BBitmap *title)
{
	delete bitmap;
	bitmap=new BBitmap(title->Bounds(),title->ColorSpace());
	memcpy(bitmap->Bits(),title->Bits(),title->BitsLength());
}
BBitmap* CBitmapField::GetBitmap(void)
{
	return bitmap;
}
void CBitmapField::SetLevel(int level)
{
	value=level;
}
int CBitmapField::GetLevel(void)
{
	return value;
}
void CBitmapField::SetVisible(bool bVisible)
{
	this->bVisible=bVisible;
}
bool CBitmapField::IsVisible(void)
{
	return bVisible;
}

CBitmapListColumn::CBitmapListColumn(BBitmap *title,float width,float minWidth,float maxWidth)
:CBitmapColumn(title,width,minWidth,maxWidth)
{
}
void CBitmapListColumn::DrawField(BView *targetView,BRect rect,CField *rowdata,bool bSelected,rgb_color highlight)
{
	BRect bumprect;

	CBitmapColumn::DrawField(targetView,rect,rowdata,bSelected,highlight);
	if(((CBitmapField*)rowdata)->bVisible) {
		targetView->SetDrawingMode(B_OP_ALPHA);
		bumprect=((CBitmapField*)rowdata)->bitmap->Bounds();
		targetView->DrawBitmapAsync(((CBitmapField*)rowdata)->bitmap,BPoint(rect.left+(((rect.right-rect.left)-(bumprect.right-bumprect.left))/2),rect.top+(((rect.bottom-rect.top)-(bumprect.bottom-bumprect.top))/2)));
	}
}
int CBitmapListColumn::CompareFields(CField *row1,CField *row2)
{
	if(((CBitmapField*)row1)->value==((CBitmapField*)row2)->value) {
		if(((CBitmapField*)row1)->bVisible==((CBitmapField*)row2)->bVisible) return 0;
		if(((CBitmapField*)row1)->bVisible) return 1;
		else return -1;
	} else
		return ((CBitmapField*)row1)->value-((CBitmapField*)row2)->value;
}
