#include <stdio.h>
#include <string.h>
#include <Bitmap.h>
#include <Directory.h>
#include <Entry.h>
#include <Mime.h>
#include <NodeInfo.h>
#include <Path.h>
#include "DirList.h"
#include "ListViewerTypes.h"

CDirectoryRow::CDirectoryRow(CColumnListView *view,const char *path)
{
	char dirname[100];

	lview=view;
	directory=new char[strlen(path)+1];
	strcpy(directory,path);
	
	icon=FindNodeIcon(directory);
	
	GetLeafName(directory,dirname);
	AddField(new CStringField((char*)((strcmp(dirname,"/")!=0)?dirname:"System")),FIELD_DIRECTORY);

	// Now configure the "Expandable" option
	BEntry entry;
	BDirectory thisdirectory(directory);
	thisdirectory.Rewind();
	bIsExpandable=false;
	while((thisdirectory.GetNextEntry(&entry,false)==B_OK)&&(!bIsExpandable)) {
		BDirectory testdirectory(&entry);
		if(testdirectory.InitCheck()==B_OK) bIsExpandable=true;
	}
}
CDirectoryRow::~CDirectoryRow()
{
	if(icon!=NULL) delete icon;
	delete[] directory;
}
bool CDirectoryRow::DrawIcon(BView *targetView,BRect rect,float *width)
{
	BRect bounds;

	if(icon!=NULL) {
		targetView->SetDrawingMode(B_OP_ALPHA);
		bounds=icon->Bounds();
		targetView->DrawBitmapAsync(icon,BPoint(rect.left,rect.top+(((rect.bottom-rect.top)-(bounds.bottom-bounds.top))/2)));
		(*width)=bounds.right-bounds.left+2;
		return true;
	} else return false;
}
bool CDirectoryRow::IsExpandable(void)
{
	return bIsExpandable;
}
void CDirectoryRow::ExpandOperation(bool bExpanding)
{
	BEntry entry;
	BDirectory thisdirectory(directory);
	CDirectoryRow *newchild;
	const char *dirname;

	if(bExpanding) {
		// We're expanding, so we must allocate a sub-item for each child we find in
		// our directory
		thisdirectory.Rewind();
		while(thisdirectory.GetNextEntry(&entry,false)==B_OK) {
			BDirectory child(&entry);
			if(child.InitCheck()==B_OK) {
				BPath path(&entry);
				dirname=path.Path();
				if(UsableDirectory(dirname)) {
					newchild=new CDirectoryRow(lview,dirname);
					lview->AddRow(newchild,this);
				}
			}
		}
	}
	CRow::ExpandOperation(bExpanding);
}
void CDirectoryRow::ExpandOperationComplete(bool bExpanding)
{
	CRow *row;

	if(!bExpanding) {
		// We're closing, so we can free the allocated sub-items to save memory
		while((row=lview->RowAt(0,NULL,this))!=NULL) {
			lview->RemoveRow(row);
			delete row;	// This also destroys all the children of the row
		}
	}
	CRow::ExpandOperationComplete(bExpanding);
}
void CDirectoryRow::GetPath(char *buffer)
{
	const char *s;

	BPath path(directory,NULL,false);
	s=path.Path();
	if(s[strlen(s)-1]!='/') sprintf(buffer,"%s/",s);
	else strcpy(buffer,s);
}
void CDirectoryRow::GetName(char *buffer)
{
	char buf[300];
	
	GetPath(buf);
	GetLeafName(buf,buffer);
}
void CDirectoryRow::GetLeafName(const char *in_buf,char *out_buf)
{
	int i,j;
	
	for(i=0;in_buf[i]!='\0';i++);
	i--;
	if(i==0) {
		out_buf[0]='/';
		out_buf[1]='\0';
		return;
	}
	if(in_buf[i]=='/') while(in_buf[i]=='/') i--;
	while(in_buf[i]!='/') i--;
	i++;
	for(j=0;(in_buf[i]!='/')&&(in_buf[i]!='\0');i++,j++) out_buf[j]=in_buf[i];
	out_buf[j]='\0';
}
bool CDirectoryRow::UsableDirectory(const char *in_dir)
{
	const char *nodirs[]={
		"/bin",
		"/dev",
		"/system",
		"/pipe",
		"/etc",
		"/tmp",
		"/var",
		NULL,
	};
	int i;
	
	for(i=0;nodirs[i]!=NULL;i++)
		if(strcmp(in_dir,nodirs[i])==0)
			return false;
	return true;
}
BBitmap* CDirectoryRow::FindNodeIcon(const char *directory)
{
	BBitmap *bitmap=new BBitmap(BRect(0,0,15,15),B_CMAP8);
	BNode node(directory);
	BNodeInfo nodeinfo(&node);
	const char *s;
	int i,j,k;
	
	if((nodeinfo.InitCheck()==B_OK)&&(nodeinfo.GetIcon(bitmap,B_MINI_ICON)==B_OK))
		return bitmap;
	else {
		if(strcmp(directory,"/")==0) s=MIME_ROOT;
		else {
			k=strlen(directory);
			for(i=0,j=0;directory[i]!='\0';i++)
				if(directory[i]=='/') j++;
			if(directory[k-1]=='/') j--;
			if((k>1)&&(j==1)) s=MIME_VOLUME;
			else s=MIME_FOLDER;
		}
		BMimeType MIMEtype(s);
		if((MIMEtype.InitCheck()==B_OK)&&(MIMEtype.GetIcon(bitmap,B_MINI_ICON)==B_OK)) return bitmap;
	}
	// All attempts to locate a bitmap failed
	delete bitmap;
	return NULL;
}
